from typing import TypedDict


class DeduplicationOptions(TypedDict, total=False):
    """
    Deduplication options.
    """

    id: str
    """
    Identifier for deduplication.
    
    Required field that uniquely identifies the deduplication key.
    """

    ttl: int
    """
    Time-to-live in milliseconds for the deduplication key.
    
    If not provided, the deduplication will last until the job is completed or failed.
    """

    extend: bool
    """
    If true, extend the TTL on each duplicate job attempt.
    
    When a duplicate job is detected, the TTL will be reset to the original value.
    """

    replace: bool
    """
    If true, replace the job data when a duplicate is added (while delayed).
    
    This is useful in debounce mode where you want to keep only the latest job data.
    """



import argparse
import importlib
import os
import sys
from unittest.mock import Mock

import setuptools

parser = argparse.ArgumentParser(description='Check if version is set inside setup.py')
parser.add_argument('setup_path', type=str, default="./setup.py", help="path of setup.py")
args = parser.parse_args()

sys.path.insert(0, os.path.dirname(args.setup_path))

mock = Mock()
setuptools.setup = mock

#import setup
importlib.import_module(os.path.splitext(os.path.basename(args.setup_path))[0])

if mock.call_args is not None and 'version' in mock.call_args[1]:
    sys.exit(f"version in {args.setup_path}")

sys.exit(0)


import argparse
import os

from tomlkit import dumps, parse

parser = argparse.ArgumentParser(description='Set version in pyproject.toml')
parser.add_argument('srcdir', type=str, help="source directory")
parser.add_argument('-v', '--version', type=str, default="1.0.0", help="version")
args = parser.parse_args()

pyproject_path = os.path.join(args.srcdir, 'pyproject.toml')

# read file
file = open(pyproject_path, "r")
pyproject = file.read()
file.close()

# modifiy version (tomlkit will preserve comments)
toml = parse(pyproject)
if 'tool' in toml and 'poetry' in toml['tool'] and "version" in toml["tool"]["poetry"]:
    print('Poetry package detected')
    toml['tool']['poetry']['version'] = args.version
else:
    toml['project']['version'] = args.version
print(f'Set version to {args.version} ({pyproject_path})')
pyproject = dumps(toml)

# write file
file = open(pyproject_path,"w")
file.write(pyproject)
file.close()





from typing import TypedDict, Any, Union
import redis.asyncio as redis
from bullmq.types.job_options import JobOptions


class QueueBaseOptions(TypedDict, total=False):
    """
    Options for the Queue class.
    """

    prefix: str
    """
    Prefix for all queue keys.
    """

    connection: Union[dict[str, Any], redis.Redis, str]
    """
    Options for connecting to a Redis instance.
    """

    defaultJobOptions: JobOptions
    """
    Default job options that will be applied to all jobs added to the queue.
    These can be overridden by individual job options.
    """



import redis.asyncio as redis
from typing import Union
from redis.backoff import ExponentialBackoff
from redis.asyncio.retry import Retry
from redis.exceptions import (
   BusyLoadingError,
   ConnectionError,
   TimeoutError
)
import warnings
import os
from bullmq.utils import isRedisVersionLowerThan, is_redis_cluster, get_cluster_nodes, get_node_client

basePath = os.path.dirname(os.path.realpath(__file__))

# Script definitions mapping script names to their file names
SCRIPT_DEFINITIONS = {
    "addStandardJob": "addStandardJob-9.lua",
    "addDelayedJob": "addDelayedJob-6.lua",
    "addParentJob": "addParentJob-6.lua",
    "addPrioritizedJob": "addPrioritizedJob-9.lua",
    "changePriority": "changePriority-7.lua",
    "cleanJobsInSet": "cleanJobsInSet-3.lua",
    "drain": "drain-5.lua",
    "extendLock": "extendLock-2.lua",
    "getCounts": "getCounts-1.lua",
    "getCountsPerPriority": "getCountsPerPriority-4.lua",
    "getRanges": "getRanges-1.lua",
    "getState": "getState-8.lua",
    "getStateV2": "getStateV2-8.lua",
    "isJobInList": "isJobInList-1.lua",
    "moveStalledJobsToWait": "moveStalledJobsToWait-8.lua",
    "moveToActive": "moveToActive-11.lua",
    "moveToDelayed": "moveToDelayed-8.lua",
    "moveToFinished": "moveToFinished-14.lua",
    "moveToWaitingChildren": "moveToWaitingChildren-7.lua",
    "obliterate": "obliterate-2.lua",
    "pause": "pause-7.lua",
    "promote": "promote-9.lua",
    "removeJob": "removeJob-2.lua",
    "reprocessJob": "reprocessJob-8.lua",
    "retryJob": "retryJob-11.lua",
    "moveJobsToWait": "moveJobsToWait-8.lua",
    "saveStacktrace": "saveStacktrace-1.lua",
    "updateData": "updateData-1.lua",
    "updateProgress": "updateProgress-3.lua",
}


def loadScript(name: str) -> str:
    """
    Load a Lua script by name from the commands directory.
    """
    with open(f"{basePath}/commands/{name}", "r") as file:
        return file.read()


class RedisConnection:
    """
    RedisConnection class
    """

    minimum_version = '5.0.0'
    recommended_minimum_version = '6.2.0'

    capabilities = {
        "canBlockFor1Ms": True,
        "canDoubleTimeout": False
    }

    def __init__(self, redisOpts: Union[dict, str, redis.Redis] = {}):
        self.version = None
        retry = Retry(ExponentialBackoff(cap=20, base=1), 20)
        retry_errors = [BusyLoadingError, ConnectionError, TimeoutError]

        if isinstance(redisOpts, redis.Redis):
            self.conn = redisOpts
        elif isinstance(redisOpts, dict):
            defaultOpts = {
                "host": "localhost",
                "port": 6379,
                "db": 0,
                "password": None,
                "username": None,
            }
            finalOpts = {**defaultOpts, **redisOpts}

            self.conn = redis.Redis(decode_responses=True, retry=retry, retry_on_error=retry_errors, **finalOpts)
        else:
            self.conn = redis.from_url(redisOpts, decode_responses=True, retry=retry,
                retry_on_error=retry_errors)

        self.commands = {}
        self.loadCommands()

    def loadCommands(self):
        """
        Load and register all Lua scripts on the Redis client.
        This is called once during initialization to avoid re-registering
        scripts on every Scripts instance creation.
        """
        for name, filename in SCRIPT_DEFINITIONS.items():
            self.commands[name] = self.conn.register_script(loadScript(filename))

    def disconnect(self):
        """
        Disconnect from Redis
        """
        return self.conn.disconnect()

    async def close(self):
        """
        Close the connection
        """
        return await self.conn.aclose()

    async def getRedisVersion(self):
        if self.version is not None:
            return self.version

        doc = await self.conn.info()
        if doc.get("maxmemory_policy") != "noeviction":
            warnings.warn(f'IMPORTANT! Eviction policy is {doc.get("maxmemory_policy")}. It should be "noeviction"')

        self.version = doc.get("redis_version")

        self.capabilities = {
            "canBlockFor1Ms": not isRedisVersionLowerThan(self.version, '7.0.8'),
            "canDoubleTimeout": not isRedisVersionLowerThan(self.version, '6.0.0')
        }
        return self.version

    async def set_client_name(self, name: str):
        if not name:
            return

        self.client_name = name

        if is_redis_cluster(self.conn):
            nodes = get_cluster_nodes(self.conn)
            for node in nodes:
                node_client = get_node_client(node)
                self._set_client_name_on_pool(node_client, name)
                await self._set_client_name_on_client(node_client, name)
        else:
            self._set_client_name_on_pool(self.conn, name)
            await self._set_client_name_on_client(self.conn, name)

    async def _set_client_name_on_client(self, client, name: str):
        if hasattr(client, "client_setname"):
            await client.client_setname(name)
        else:
            await client.execute_command("CLIENT", "SETNAME", name)

    def _set_client_name_on_pool(self, client, name: str):
        pool = getattr(client, "connection_pool", None)
        if pool is None:
            return

        connection_kwargs = getattr(pool, "connection_kwargs", None)
        if isinstance(connection_kwargs, dict):
            connection_kwargs["client_name"] = name



import json
import traceback
from typing import Any

import semver


def isRedisVersionLowerThan(current_version, minimum_version):
    return semver.Version.parse(current_version).compare(minimum_version) == -1


def extract_result(job_task, emit_callback):
    try:
        return job_task.result()
    except Exception as e:
        if not str(e).startswith('Connection closed by server'):
            # lets use a simple-but-effective error handling:
            # ignore the job
            traceback.print_exc()
            emit_callback("error", e)

def get_parent_key(opts: dict[str, str]):
    if opts:
        return f"{opts.get('queue')}:{opts.get('id')}"

def parse_json_string_values(input_dict: dict[str, str]) -> dict[str, dict]:
    return {key: json.loads(value) for key, value in input_dict.items()}

def object_to_flat_array(obj: dict[str, Any]) -> list[Any]:
    """
    Converts a dictionary into a flat array where each key is followed by its value.

    Args:
        obj (dict[str, Any]): The input dictionary to flatten.

    Returns:
        list[Any]: A flat list containing keys and values from the dictionary in order.
    """
    arr = []
    for key, value in obj.items():
        arr.append(key)
        arr.append(value)
    return arr


def is_redis_cluster(client: Any) -> bool:
    try:
        from redis.asyncio.cluster import RedisCluster
    except Exception:
        RedisCluster = None

    if RedisCluster is not None and isinstance(client, RedisCluster):
        return True

    return bool(
        getattr(client, "is_cluster", False)
        or getattr(client, "isCluster", False)
        or getattr(client, "nodes_manager", None)
    )


def get_cluster_nodes(client: Any) -> list[Any]:
    if hasattr(client, "get_nodes") and callable(getattr(client, "get_nodes")):
        try:
            return list(client.get_nodes())
        except Exception:
            return []

    nodes_manager = getattr(client, "nodes_manager", None)
    if nodes_manager is not None and hasattr(nodes_manager, "nodes_cache"):
        try:
            return list(nodes_manager.nodes_cache.values())
        except Exception:
            return []

    if hasattr(client, "nodes") and callable(getattr(client, "nodes")):
        try:
            return list(client.nodes())
        except Exception:
            return []

    return []


def get_node_client(node: Any) -> Any:
    for attr in ("client", "redis", "connection", "redis_connection"):
        if hasattr(node, attr):
            return getattr(node, attr)
    return node
